
/* -------------------------------- Include Directives -------------------------------- */
/* HAL */

#include "api_ble.h"

LOG_MODULE_REGISTER(api_ble, 4);

/* --------------------------------- Global Variables --------------------------------- */


struct bt_conn *default_conn;
struct bt_le_conn_param *conn_param = BT_LE_CONN_PARAM( CONFIG_INTERVAL_MIN, 
                                                        CONFIG_INTERVAL_MAX, 
                                                        CONFIG_CONNECTION_LATENCY, 
                                                        CONFIG_SUPERVISION_TIMEOUT);

BT_CONN_CTX_DEF(conns, CONFIG_BT_MAX_CONN, sizeof(struct bt_nus_client));

BT_SCAN_CB_INIT(scan_cb,
                ifeel_on_scan_filter_match,
                ifeel_on_scan_filter_no_match,
                ifeel_on_scan_connecting_error, 
				ifeel_on_scan_connecting);

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = ifeel_on_bt_connected,
	.disconnected = ifeel_on_bt_disconnected
};

extern struct bt_gatt_dm_cb discovery_cb = {
	.completed         = ifeel_on_discovery_complete,
	.service_not_found = ifeel_on_discovery_service_not_found,
	.error_found       = ifeel_on_discovery_error,
};

extern struct bt_conn_auth_cb conn_auth_callbacks = {
	.cancel = auth_cancel,
	.pairing_confirm = pairing_confirm
};

extern struct bt_conn_auth_info_cb conn_auth_info_callbacks = {
	.pairing_complete = pairing_complete,
	.pairing_failed = pairing_failed
};

/* -------------------------------- Function Definitions -------------------------------*/


ifeel_t ifeel_ble_init(void){

	int err;

	err = bt_conn_auth_cb_register(&conn_auth_callbacks);
	if (err) {
		LOG_ERR("Failed to register authorization callbacks.");
		return 0;
	}

	err = bt_conn_auth_info_cb_register(&conn_auth_info_callbacks);
	if (err) {
		printk("Failed to register authorization info callbacks.\n");
		return 0;
	}

	err = bt_enable(NULL);
	if (err) {
		LOG_INF("Bluetooth init failed (err %d)", err);
		return err;
	}
	if (IS_ENABLED(CONFIG_SETTINGS)) {
		settings_load();
	}

    iFEEL(ifeel_ble_scan_init());
    iFEEL(ifeel_ble_scan_start());

}

ifeel_t ifeel_ble_scan_start(void){

    int err;
	err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	if (err) {
		LOG_INF("Starting scanning failed (err %d)", err);
		return err;
	}
    LOG_INF("BLE Scanning started");
    return 0;
 }



ifeel_t ifeel_ble_scan_init(void){
    int err;
	struct bt_le_scan_param scan_param = {
		.type = BT_LE_SCAN_TYPE_ACTIVE,
		.options = BT_LE_SCAN_OPT_FILTER_DUPLICATE,
		.interval = 0x0010,
		.window = 0x0010,
	};

	struct bt_scan_init_param scan_init = {
		.connect_if_match = 1,
		//.scan_param = &scan_param,
		//.conn_param = conn_param
	};

	bt_scan_init(&scan_init);
	bt_scan_cb_register(&scan_cb);

	// err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_NAME, "iFeel-Peripheral");
	// if (err) {
	// 	LOG_INF("Scanning filters cannot be set");
	// 	return err;
	// }

    // err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_NAME, "iFeel-Peripheral1");
	// if (err) {
	// 	LOG_INF("Scanning filters cannot be set");
	// 	return err;
	// }

	// err = bt_scan_filter_enable(BT_SCAN_NAME_FILTER, false);
	// if (err) {
	// 	LOG_INF("Filters cannot be turned on");
    //     return err;
	// }

	err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_UUID, BT_UUID_NUS_SERVICE);
	if (err) {
		LOG_ERR("Scanning filters cannot be set (err %d)", err);
		return err;
	}

	err = bt_scan_filter_enable(BT_SCAN_UUID_FILTER, false);
	if (err) {
		LOG_ERR("Filters cannot be turned on (err %d)", err);
		return err;
	}
    
    LOG_INF("BLE Scanning initialized");

    return 0;
}

void auth_cancel(struct bt_conn *conn)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Pairing cancelled: %s", addr);
}


void pairing_confirm(struct bt_conn *conn)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	bt_conn_auth_pairing_confirm(conn);

	LOG_INF("Pairing confirmed: %s", addr);
}


void pairing_complete(struct bt_conn *conn, bool bonded)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Pairing completed: %s, bonded: %d", addr,bonded);
}

void pairing_failed(struct bt_conn *conn, enum bt_security_err reason)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_WRN("Pairing failed conn: %s, reason %d", addr,	reason);
}


void ifeel_on_scan_filter_match(struct bt_scan_device_info *device_info, bool connectable){

	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(device_info->recv_info->addr, addr, sizeof(addr));

	LOG_INF("Filters matched. Address: %s connectable: %d",
		addr, connectable);
}

void ifeel_on_scan_filter_no_match(struct bt_scan_device_info *device_info, bool connectable)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(device_info->recv_info->addr, addr, sizeof(addr));

	//LOG_INF("Filter not match. Address: %s connectable: %d", addr, connectable);
}

void ifeel_on_scan_connecting_error(struct bt_scan_device_info *device_info){

	LOG_INF("Connecting failed");
	
	ifeel_ble_scan_start();
}

void ifeel_on_scan_connecting(struct bt_scan_device_info *device_info,
			    struct bt_conn *conn)
{
	default_conn = bt_conn_ref(conn);
}

void ifeel_on_discovery_complete(struct bt_gatt_dm *dm, void *context){
    
    int err;
	struct bt_nus_client *nus = context;

	struct bt_throughput *throughput = context;
    LOG_INF("Service discovery complete");
    bt_gatt_dm_data_print(dm);

	// bt_nus_handles_assign(dm, nus);
	// bt_nus_subscribe_receive(nus);

	// bt_gatt_dm_data_release(dm);
	ifeel_ble_scan_start();


}

void ifeel_on_discovery_service_not_found(struct bt_conn *conn, void *context){
    LOG_INF("Service not found");
}

void ifeel_on_discovery_error(struct bt_conn *conn, int err, void *context){
    LOG_INF("Discovery error: %d", err);
}

void ifeel_on_bt_connected(struct bt_conn *conn, uint8_t hci_err){

	char addr[BT_ADDR_LE_STR_LEN];
    struct bt_conn_info info = {0};
	
	int err;

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (hci_err) {

		LOG_INF("Connection failed (err 0x%02x)", hci_err);
		if (default_conn == conn) {
			bt_conn_unref(default_conn);
			default_conn = NULL;

			ifeel_ble_scan_start();
		}
		return;
	}
    
	default_conn = bt_conn_ref(conn);

	LOG_INF("----- Connected to %s -----", addr);

    // err = ifeel_ble_conn_set(BT_LE_CONN_PARAM(CONFIG_INTERVAL_MIN,
    //                         CONFIG_INTERVAL_MAX,
    //                         CONFIG_CONNECTION_LATENCY, CONFIG_SUPERVISION_TIMEOUT),
    //                         BT_CONN_LE_PHY_PARAM_2M,
    //                         BT_LE_DATA_LEN_PARAM_MAX);
    // if (err) {
    //     LOG_INF("Connection parameters set failed (err %d)", err);
    // }

	struct bt_nus_client *nus_client = bt_conn_ctx_alloc(&conns_ctx_lib, conn);
	if (!nus_client) {
		LOG_WRN("There is no free memory to "
			"allocate the connection context");
	}
	
	struct bt_nus_client_init_param nus_client_param = {
		.cb = {
			.received = NULL,
			.sent = NULL,
		}
	};

	memset(nus_client, 0, bt_conn_ctx_block_size_get(&conns_ctx_lib));

	err = bt_nus_client_init(nus_client, &nus_client_param);

    err = bt_gatt_dm_start(default_conn,
                    BT_UUID_NUS_SERVICE,
                    &discovery_cb,
                    &nus_client);

	if (err) {
		LOG_ERR("could not start the discovery procedure, error "
			"code: %d", err);
	}
	
	bt_conn_ctx_release(&conns_ctx_lib, (void *)nus_client);

	err = bt_scan_stop();
	if ((!err) && (err != -EALREADY)) {
		LOG_ERR("Stop LE scan failed (err %d)", err);
	}	

}

void ifeel_on_bt_disconnected(struct bt_conn *conn, uint8_t reason){
	
	int err;

    LOG_INF("Disconnected (reason %u)", reason);
    
	err = bt_conn_ctx_free(&conns_ctx_lib, conn);

	if (err) {
		LOG_WRN("The memory was not allocated for the context of this "
			"connection.");
	}
    
	bt_conn_unref(default_conn);
	default_conn = NULL;

	ifeel_ble_scan_start();
}

ifeel_t ifeel_ble_conn_set(const struct bt_le_conn_param *conn_param,
			const struct bt_conn_le_phy_param *phy,
			const struct bt_conn_le_data_len_param *data_len){

    int err;
	struct bt_conn_info info = {0};

	err = bt_conn_get_info(default_conn, &info);
	if (err) {
		LOG_INF("Failed to get connection info %d", err);
		return err;
	}

	err = bt_conn_le_phy_update(default_conn, phy);
	if (err) {
		LOG_INF("PHY update failed: %d", err);
		return err;
	}

	if (err) {
		LOG_INF("PHY update timeout");
		return err;
	}

	if (info.le.data_len->tx_max_len != data_len->tx_max_len) {

		err = bt_conn_le_data_len_update(default_conn, data_len);
		if (err) {
			LOG_INF("LE data length update failed: %d",
				    err);
			return err;
		}

		LOG_INF("LE Data length updated");
		if (err) {
			LOG_INF("LE Data Length update timeout: %d", err);
			return err;
		}
	}

	if (info.le.interval != conn_param->interval_max) {
		err = bt_conn_le_param_update(default_conn, conn_param);
		if (err) {
			LOG_INF("Connection parameters update failed: %d", err);
			return err;
		}

		LOG_INF("Connection parameters updated");
		if (err) {
			LOG_INF("Connection parameters update timeout: %d", err);
			return err;
		}
	}

	return 0;
}

/* End of file FILENAME */
